# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
# pylint: disable=too-many-lines

from __future__ import annotations

import re
from abc import ABC
from collections.abc import Callable, Iterable, Sequence
from contextlib import contextmanager
from io import BytesIO
from json import dumps, loads
from pathlib import Path
from typing import TYPE_CHECKING, Any
from uuid import NAMESPACE_DNS, UUID, uuid5
from warnings import warn

import h5py
import numpy as np

from .exceptions import Geoh5FileClosedError


if TYPE_CHECKING:
    from ..workspace import Workspace
    from .entity import Entity
    from .entity_container import EntityContainer

INV_KEY_MAP = {
    "Allow delete": "allow_delete",
    "Allow delete contents": "allow_delete_content",
    "Allow move": "allow_move",
    "Allow move contents": "allow_move_content",
    "Allow rename": "allow_rename",
    "Association": "association",
    "Attributes": "concatenated_attributes",
    "Blob": "BLOB",
    "Boolean": "BOOLEAN",
    "Cell": "CELL",
    "Cells": "cells",
    "Clipping IDs": "clipping_ids: list | None",
    "Collar": "collar",
    "Color map": "color_map",
    "Colour": "COLOUR",
    "Contributors": "contributors",
    "Concatenated object IDs": "concatenated_object_ids",
    "Cost": "cost",
    "Current line property ID": "current_line_id",
    "Data": "values",
    "DateTime": "DATETIME",
    "Description": "description",
    "Dip": "dip",
    "Distance unit": "distance_unit",
    "Dynamic implementation ID": "dynamic_implementation_id",
    "Duplicate type on copy": "duplicate_type_on_copy",
    "End of hole": "end_of_hole",
    "Face": "FACE",
    "File name": "name",
    "Filename": "FILENAME",
    "Float": "FLOAT",
    "Geometric": "GEOMETRIC",
    "Group": "GROUP",
    "Group Name": "name",
    "GA Version": "ga_version",
    "Hidden": "hidden",
    "Invalid": "INVALID",
    "Integer": "INTEGER",
    "ID": "uid",
    "Last focus": "last_focus",
    "Layers": "layers",
    "Mapping": "mapping",
    "Metadata": "metadata",
    "Modifiable": "modifiable",
    "Multi-Text": "MULTI_TEXT",
    "Name": "name",
    "Number of bins": "number_of_bins",
    "NU": "u_count",
    "Nu": "u_count",
    "NV": "v_count",
    "Nv": "v_count",
    "NW": "w_count",
    "options": "options",
    "Object": "OBJECT",
    "Origin": "origin",
    "Octree Cells": "octree_cells",
    "Partially hidden": "partially_hidden",
    "Planning": "planning",
    "Precision": "precision",
    "Primitive type": "primitive_type",
    "Prisms": "prisms",
    "Properties": "properties",
    "Property Group IDs": "property_group_ids",
    "Property Group Type": "property_group_type",
    "PropertyGroups": "property_groups",
    "Public": "public",
    "Referenced": "REFERENCED",
    "Rotation": "rotation",
    "Scale": "scale",
    "Scientific notation": "scientific_notation",
    "Surveys": "surveys",
    "Text": "TEXT",
    "TextMesh Data": "text_mesh_data",
    "Trace": "trace",
    "TraceDepth": "trace_depth",
    "Transparent no data": "transparent_no_data",
    "Unknown": "UNKNOWN",
    "U cell delimiters": "u_cell_delimiters",
    "V cell delimiters": "v_cell_delimiters",
    "Z cell delimiters": "z_cell_delimiters",
    "U Cell Size": "u_cell_size",
    "U Count": "u_count",
    "U Size": "u_cell_size",
    "U size": "u_cell_size",
    "V Cell Size": "v_cell_size",
    "V Count": "v_count",
    "V Size": "v_cell_size",
    "V size": "v_cell_size",
    "Vector": "VECTOR",
    "Version": "version",
    "Vertical": "vertical",
    "Vertices": "vertices",
    "Vertex": "VERTEX",
    "Visible": "visible",
    "W Cell Size": "w_cell_size",
    "Flag property ID": "flag_property_id",
    "Heterogeneous property ID": "heterogeneous_property_id",
    "Physical data name": "physical_data_name",
    "Unit property ID": "unit_property_id",
    "Weight property ID": "weight_property_id",
}

KEY_MAP = {value: key for key, value in INV_KEY_MAP.items()}

PNG_KWARGS = {"format": "PNG", "compress_level": 9}
JPG_KWARGS = {"format": "JPEG", "quality": 85}
TIF_KWARGS = {"format": "TIFF"}

PILLOW_ARGUMENTS = {
    "1": PNG_KWARGS,
    "L": PNG_KWARGS,
    "P": PNG_KWARGS,
    "RGB": PNG_KWARGS,
    "RGBA": PNG_KWARGS,
    "CMYK": JPG_KWARGS,
    "YCbCr": JPG_KWARGS,
    "I": TIF_KWARGS,
    "F": TIF_KWARGS,
}


def copy_no_reference(values: dict) -> dict:
    """
    Copy a dictionary without references to objects UUID.

    :param values: The dictionary to copy.

    :return: The copied dictionary.
    """
    # Copy metadata except reference to entities UUID
    output = {}
    for key, value in values.items():
        if isinstance(value, dict):
            value = copy_no_reference(value)

        if isinstance(value, UUID):
            value = None

        output[key] = value

    return output


@contextmanager
def fetch_active_workspace(workspace: Workspace | None, mode: str = "r"):
    """
    Open a workspace in the requested 'mode'.

    If receiving an opened Workspace instead, merely return the given workspace.

    :param workspace: A Workspace class
    :param mode: Set the h5 read/write mode

    :return h5py.File: Handle to an opened Workspace.
    """
    try:
        geoh5 = None if workspace is None else workspace.geoh5
    except Geoh5FileClosedError:
        geoh5 = None

    if workspace is None or (geoh5 is not None and mode in workspace.geoh5.mode):
        try:
            yield workspace
        finally:
            pass
    else:
        if geoh5 is not None:
            warn(
                f"Closing the workspace in mode '{workspace.geoh5.mode}' "
                f"and re-opening in mode '{mode}'."
            )
            workspace.close()

        try:
            yield workspace.open(mode=mode)
        finally:
            workspace.close()


@contextmanager
def fetch_h5_handle(file: str | h5py.File | Path, mode: str = "r") -> h5py.File:
    """
    Open in read+ mode a geoh5 file from string.
    If receiving a file instead of a string, merely return the given file.

    :param file: Name or handle to a geoh5 file.
    :param mode: Set the h5 read/write mode

    :return h5py.File: Handle to an opened h5py file.
    """
    if isinstance(file, h5py.File):
        try:
            yield file
        finally:
            pass
    else:
        if Path(file).suffix != ".geoh5":
            raise ValueError("Input h5 file must have a 'geoh5' extension.")

        h5file = h5py.File(file, mode)

        try:
            yield h5file
        finally:
            h5file.close()


def match_values(vec_a, vec_b, collocation_distance=1e-4) -> np.ndarray:
    """
    Find indices of matching values between two arrays, within collocation_distance.

    :param: vec_a, list or numpy.ndarray
        Input sorted values

    :param: vec_b, list or numpy.ndarray
        Query values

    :return: indices, numpy.ndarray
        Pairs of indices for matching values between the two arrays such
        that vec_a[ind[:, 0]] == vec_b[ind[:, 1]].
    """
    ind_sort = np.argsort(vec_a)
    ind = np.minimum(
        np.searchsorted(vec_a[ind_sort], vec_b, side="right"), vec_a.shape[0] - 1
    )
    nearests = np.c_[ind, ind - 1]
    match = np.where(
        np.abs(vec_a[ind_sort][nearests] - vec_b[:, None]) < collocation_distance
    )
    indices = np.c_[ind_sort[nearests[match[0], match[1]]], match[0]]
    return indices


def merge_arrays(
    head,
    tail,
    *,
    replace="A->B",
    mapping=None,
    collocation_distance=1e-4,
    return_mapping=False,
) -> np.ndarray:
    """
    Given two numpy.arrays of different length, find the matching values and append both arrays.

    :param: head, numpy.array of float
        First vector of shape(M,) to be appended.
    :param: tail, numpy.array of float
        Second vector of shape(N,) to be appended
    :param: mapping=None, numpy.ndarray of int
        Optional array where values from the head are replaced by the tail.
    :param: collocation_distance=1e-4, float
        Tolerance between matching values.

    :return: numpy.array shape(O,)
        Unique values from head to tail without repeats, within collocation_distance.
    """

    if mapping is None:
        mapping = match_values(head, tail, collocation_distance=collocation_distance)

    if mapping.shape[0] > 0:
        if replace == "B->A":
            head[mapping[:, 0]] = tail[mapping[:, 1]]
        else:
            tail[mapping[:, 1]] = head[mapping[:, 0]]

        tail = np.delete(tail, mapping[:, 1])

    if return_mapping:
        return np.r_[head, tail], mapping

    return np.r_[head, tail]


def clear_array_attributes(entity: Entity, recursive: bool = False):
    """
    Clear all stashed values of attributes from an entity to free up memory.

    :param entity: Entity to clear attributes from.
    :param recursive: Clear attributes from children entities.
    """
    if isinstance(entity.workspace.h5file, BytesIO):
        return

    for attribute in [
        "vertices",
        "cells",
        "values",
        "prisms",
        "layers",
        "octree_cells",
    ]:
        if hasattr(entity, attribute):
            setattr(entity, f"_{attribute}", None)

    if recursive and hasattr(entity, "children"):
        for child in entity.children:
            clear_array_attributes(child, recursive=recursive)


def are_objects_similar(obj1, obj2, ignore: list[str] | None):
    """
    Compare two objects to see if they are similar. This is a shallow comparison.

    :param obj1: The first object.
    :param obj2: The first object.
    :param ignore: List of attributes to ignore.

    :return: If attributes similar or not.
    """
    if not isinstance(obj1, type(obj2)):
        raise TypeError("Objects are not the same type.")

    attributes1 = getattr(obj1, "__dict__", obj1)
    attributes2 = getattr(obj2, "__dict__", obj2)

    # remove the ignore attributes
    if isinstance(ignore, list) and isinstance(attributes1, dict):
        for item in ignore:
            attributes1.pop(item, None)
            attributes2.pop(item, None)

    return attributes1 == attributes2


def compare_arrays(object_a, object_b, attribute: str, decimal: int = 6):
    """
    Utility to compare array properties from two Entities

    :param object_a: First Entity
    :param object_b: Second Entity
    :param attribute: Attribute to compare
    :param decimal: Decimal precision for comparison
    """
    array_a_values = getattr(object_a, attribute)
    array_b_values = getattr(object_b, attribute)

    if array_b_values is None:
        raise ValueError(f"attr {attribute} is None for object {object_b.name}")

    if array_b_values.dtype.names is not None:
        assert all(
            np.all(array_a_values[name] == array_b_values[name])
            for name in array_b_values.dtype.names
        ), f"Error comparing attribute '{attribute}'."

    elif len(array_a_values) > 0 and isinstance(array_a_values[0], str):
        assert all(array_a_values == array_b_values), (
            f"Error comparing attribute '{attribute}'."
        )
    else:
        np.testing.assert_array_almost_equal(
            array_a_values,
            array_b_values,
            decimal=decimal,
            err_msg=f"Error comparing attribute '{attribute}'.",
        )


def compare_floats(object_a, object_b, attribute: str, decimal: int = 6):
    np.testing.assert_almost_equal(
        getattr(object_a, attribute),
        getattr(object_b, attribute),
        decimal=decimal,
        err_msg=f"Error comparing attribute '{attribute}'.",
    )


def compare_list(object_a, object_b, attribute: str, ignore: list[str] | None):
    get_object_a = getattr(object_a, attribute)
    get_object_b = getattr(object_b, attribute)
    assert isinstance(get_object_a, list)
    assert len(get_object_a) == len(get_object_b)
    for obj_a, obj_b in zip(get_object_a, get_object_b, strict=False):
        assert are_objects_similar(obj_a, obj_b, ignore)


def compare_bytes(object_a, object_b):
    assert object_a == object_b, (
        f"{type(object_a)} objects: {object_a}, {object_b} are not equal."
    )


def compare_entities(
    object_a, object_b, ignore: list[str] | None = None, decimal: int = 6
) -> None:
    if isinstance(object_a, bytes):
        compare_bytes(object_a, object_b)
        return

    base_ignore = [
        "_workspace",
        "_children",
        "_visual_parameters",
        "_entity_class",
        "_intervals",
    ]
    ignore_list = base_ignore + ignore if ignore else base_ignore

    for attr in [k for k in object_a.__dict__ if k not in ignore_list]:
        if isinstance(getattr(object_a, attr.lstrip("_")), ABC):
            compare_entities(
                getattr(object_a, attr.lstrip("_")),
                getattr(object_b, attr.lstrip("_")),
                ignore=ignore,
                decimal=decimal,
            )
        else:
            if isinstance(getattr(object_a, attr.lstrip("_")), np.ndarray):
                compare_arrays(object_a, object_b, attr.lstrip("_"), decimal=decimal)
            elif isinstance(getattr(object_a, attr.lstrip("_")), float):
                compare_floats(object_a, object_b, attr.lstrip("_"), decimal=decimal)
            elif isinstance(getattr(object_a, attr.lstrip("_")), list):
                compare_list(object_a, object_b, attr.lstrip("_"), ignore)
            else:
                try:
                    assert np.all(
                        getattr(object_a, attr.lstrip("_"))
                        == getattr(object_b, attr.lstrip("_"))
                    ), (
                        f"Output attribute '{attr.lstrip('_')}' for {object_a} do "
                        f"not match input {object_b}"
                    )
                except AssertionError:
                    pass


def is_uuid(value: str) -> bool:
    """Check if a string is UUID compliant."""
    try:
        UUID(str(value))
        return True
    except ValueError:
        return False


def entity2uuid(value: Any) -> UUID | Any:
    """Convert an entity to its UUID."""
    if hasattr(value, "uid"):
        return value.uid
    return value


def uuid2entity(value: UUID, workspace: Workspace) -> Entity | Any:
    """Convert UUID to a known entity."""
    if isinstance(value, UUID):
        if value in workspace.list_entities_name:
            return workspace.get_entity(value)[0]

        # Search for property groups
        for obj in workspace.objects:
            if obj.property_groups is not None:
                prop_group = [
                    prop_group
                    for prop_group in obj.property_groups
                    if prop_group.uid == value
                ]

                if prop_group:
                    return prop_group[0]

        return None

    return value


def str2uuid(value: Any) -> UUID | Any:
    """Convert string to UUID"""
    if isinstance(value, bytes):
        value = value.decode("utf-8")

    if is_uuid(value):
        # TODO insert validation
        return UUID(str(value))
    return value


def as_str_if_uuid(value: UUID | Any) -> str | Any:
    """Convert :obj:`UUID` to string used in geoh5."""
    if isinstance(value, UUID):
        return "{" + str(value) + "}"
    return value


def bool_value(value: np.int8) -> bool:
    """Convert logical int8 to bool."""
    return bool(value)


def as_str_if_utf8_bytes(value) -> str:
    """Convert bytes to string"""
    if isinstance(value, bytes):
        value = value.decode("utf-8")
    return value


def as_float_if_isnumeric(value: str) -> float | str:
    """Convert bytes to string"""
    if value.isnumeric():
        return float(value)
    return value


def str_json_to_dict(string: str | bytes) -> dict:
    """
    Convert a json string or bytes to a dictionary.

    :param string: The json string or bytes to convert to a dictionary.

    :return: The dictionary representation of the json string with uuid promoted.
    """
    value = as_str_if_utf8_bytes(string)
    json_dict = loads(value)

    for key, val in json_dict.items():
        if isinstance(val, dict):
            for sub_key, sub_val in val.items():
                json_dict[key][sub_key] = str2uuid(sub_val)
        else:
            json_dict[key] = str2uuid(val)

    return json_dict


def ensure_uuid(value: UUID | str) -> UUID:
    """
    Ensure that the value is a UUID.

    If not, it raises a type error.

    :param value: The value to ensure is a UUID.

    :return: The verified UUID.
    """
    value = str2uuid(value)

    if not isinstance(value, UUID):
        raise TypeError(f"Value {value} is not a UUID but a {type(value)}.")

    return value


def dict_mapper(val, string_funcs: list[Callable], *args, omit: dict | None = None):
    """
    Recursion through nested dictionaries and applies mapping functions to values.

    :param val: Value (could be another dictionary) to apply transform functions.
    :param string_funcs: Functions to apply on values within the input dictionary.
    :param omit: Dictionary of functions to omit.

    :return val: Transformed values
    """
    if isinstance(val, dict):
        val = val.copy()
        for key, values in val.items():
            short_list = string_funcs.copy()
            if omit is not None:
                short_list = [
                    fun for fun in string_funcs if fun not in omit.get(key, [])
                ]

            val[key] = dict_mapper(values, short_list)

    if isinstance(val, (list, tuple)):
        return [dict_mapper(elem, string_funcs) for elem in val]

    for fun in string_funcs:
        val = fun(val, *args)

    return val


def box_intersect(
    extent_a: np.ndarray | Sequence, extent_b: np.ndarray | Sequence
) -> bool:
    """
    Compute the intersection of two axis-aligned bounding extents defined by their
    arrays of minimum and maximum bounds in N-D space.

    :param extent_a: First extent coordinated, array or list of shape (2, N)
    :param extent_b: Second extent coordinated, array or list of shape (2, N)

    :return: Logic if the box extents intersect along all dimensions.
    """
    if isinstance(extent_a, Sequence):
        extent_a = np.vstack(extent_a)

    if isinstance(extent_b, Sequence):
        extent_b = np.vstack(extent_b)

    for extent in [extent_a, extent_b]:
        if not isinstance(extent, np.ndarray) or extent.ndim != 2:
            raise TypeError("Input extents must be 2D numpy.ndarrays.")

        if extent.shape[0] != 2 or not np.all(extent[0, :] <= extent[1, :]):
            raise ValueError(
                "Extents must be of shape (2, N) containing the minimum and maximum "
                "bounds in nd-space on the first and second row respectively."
            )

    for comp_a, comp_b in zip(extent_a.T, extent_b.T, strict=False):
        min_ext = max(comp_a[0], comp_b[0])
        max_ext = min(comp_a[1], comp_b[1])

        if min_ext > max_ext:
            return False

    return True


def mask_by_extent(
    locations: np.ndarray, extent: np.ndarray | Sequence, inverse: bool = False
) -> np.ndarray:
    """
    Find indices of locations within a rectangular extent.

    :param locations: shape(*, 3) or shape(*, 2) Coordinates to be evaluated.
    :param extent: shape(2, 2) Limits defined by the South-West and
        North-East corners. Extents can also be provided as 3D coordinates
        with shape(2, 3) defining the top and bottom limits.
    :param inverse: Return the complement of the mask extent.

    :returns: Array of bool for the locations inside or outside the box extent.
    """
    if isinstance(extent, Sequence):
        extent = np.vstack(extent)

    if not isinstance(extent, np.ndarray) or extent.ndim != 2:
        raise ValueError("Input 'extent' must be a 2D array-like.")

    if not isinstance(locations, np.ndarray) or locations.ndim != 2:
        raise ValueError(
            "Input 'locations' must be an array-like of shape(*, 3) or (*, 2)."
        )

    indices = np.ones(locations.shape[0], dtype=bool)
    for loc, lim in zip(locations.T, extent.T, strict=False):
        indices &= (lim[0] <= loc) & (loc <= lim[1])

    if inverse:
        return ~indices

    return indices


def get_attributes(entity, omit_list=(), attributes=None) -> dict:
    """Extract the attributes of an object with omissions."""
    if attributes is None:
        attributes = {}
    for key in vars(entity):
        if key not in omit_list:
            key = key.lstrip("_")
            attr = getattr(entity, key)
            attributes[key] = attr

    return attributes


def dip_azimuth_to_vector(
    dip: float | np.ndarray, azimuth: float | np.ndarray
) -> np.ndarray:
    """
    Convert dip and azimuth to a unit vector.

    :param dip: The dip angle in degree from horizontal (positive up).
    :param azimuth: The azimuth angle in degree from North (clockwise).

    :return: The unit vector.
    """
    azimuth = np.deg2rad(90 - azimuth)
    dip = np.deg2rad(dip)

    return np.c_[
        np.cos(dip) * np.cos(azimuth),
        np.cos(dip) * np.sin(azimuth),
        np.sin(dip),
    ]


def xy_rotation_matrix(angle: float) -> np.ndarray:
    """
    Rotation matrix about the z-axis.

    :param angle: Rotation angle in radians.

    :return rot: Rotation matrix.
    """
    return np.array(
        [
            [np.cos(angle), -np.sin(angle), 0.0],
            [np.sin(angle), np.cos(angle), 0.0],
            [0.0, 0.0, 1.0],
        ]
    )


def yz_rotation_matrix(angle: float) -> np.ndarray:
    """
    Rotation matrix about the x-axis.
    :param angle: Rotation angle in radians.
    :return: rot: Rotation matrix.
    """

    return np.array(
        [
            [1, 0, 0],
            [0, np.cos(angle), -np.sin(angle)],
            [0, np.sin(angle), np.cos(angle)],
        ]
    )


def dip_points(points: np.ndarray, dip: float, rotation: float = 0) -> np.ndarray:
    """
    Rotate points about the x-axis by the dip angle and then about the z-axis by the rotation angle.
    :param points: an array of points to rotate
    :param dip: the dip angle in radians
    :param rotation: the rotation angle in radians
    :return: the rotated points
    """
    # Assert points is a numpy array containing 3D points
    if not isinstance(points, np.ndarray) and points.ndim != 2 and points.shape[1] != 3:
        raise TypeError("Input points must be a 2D numpy array of shape (N, 3).")

    # rotate the points about the z-axis by the inverse rotation angle
    points = xy_rotation_matrix(-rotation) @ points.T

    # Rotate points with the dip angle
    points = yz_rotation_matrix(dip) @ points

    # Rotate back the points to initial orientation
    points = xy_rotation_matrix(rotation) @ points

    return points.T


def set_attributes(entity, **kwargs):
    """
    Loop over kwargs and set attributes to an entity.

    TODO: Deprecate in favor of explicit attribute setting.
    """
    for key, value in kwargs.items():
        try:
            setattr(entity, key, value)
        except AttributeError:
            continue


def map_name_attributes(object_, **kwargs: dict) -> dict:
    """
    Map attributes to an object. The object must have an '_attribute_map'.

    :param object_: The object to map the attributes to.
    :param kwargs: Dictionary of attributes.
    """
    mapping = getattr(object_, "_attribute_map", None)

    if mapping is None:
        raise AttributeError("Object must have an '_attribute_map' attribute.")

    new_args = {}
    for attr, item in kwargs.items():
        if attr in mapping:
            new_args[mapping[attr]] = item
        else:
            new_args[attr] = item

    return new_args


def map_attributes(object_, **kwargs):
    """
    Map attributes to an object. The object must have an '_attribute_map'.

    :param entity: The object to map the attributes to.
    :param kwargs: The kwargs to map to the object.
    """

    values = map_name_attributes(object_, **kwargs)  # Swap duplicates
    set_attributes(object_, **values)


def stringify(values: dict[str, Any]) -> dict[str, Any]:
    """
    Convert all values in a dictionary to string.

    :param values: Dictionary of values to be converted.

    :return: Dictionary of string values.
    """
    mappers = [
        entity2uuid,
        nan2str,
        inf2str,
        as_str_if_uuid,
        none2str,
        workspace2path,
        path2str,
    ]
    return dict_mapper(values, mappers)


def to_list(value: Any) -> list:
    """
    Convert value to a list.

    :param value: The value to convert.

    :return: A list
    """
    # ensure the names are a list
    if isinstance(value, list):
        return value
    if isinstance(value, tuple):
        return list(value)
    return [value]


def to_tuple(value: Any) -> tuple:
    """
    Convert value to a tuple.

    :param value: The value to convert.

    :return: A tuple
    """
    # ensure the names are a tuple
    if isinstance(value, tuple):
        return value
    if isinstance(value, list):
        return tuple(value)
    return (value,)


class SetDict(dict):
    def __init__(self, **kwargs):
        kwargs = {k: self.make_set(v) for k, v in kwargs.items()}
        super().__init__(kwargs)

    def make_set(self, value):
        if isinstance(value, (set, tuple, list)):
            value = set(value)
        else:
            value = {value}
        return value

    def __setitem__(self, key, value):
        value = self.make_set(value)
        super().__setitem__(key, value)

    def update(self, value: dict, **kwargs) -> None:  # type: ignore
        for key, val in value.items():
            val = self.make_set(val)
            if key in self:
                val = self[key].union(val)
            value[key] = val
        super().update(value, **kwargs)


def inf2str(value):  # map np.inf to "inf"
    if not isinstance(value, (int, float)):
        return value
    return str(value) if not np.isfinite(value) else value


def list2str(value):
    if isinstance(value, list):  # & (key not in exclude):
        return str(value)[1:-1]
    return value


def none2str(value):
    if value is None:
        return ""
    return value


def path2str(value):
    if isinstance(value, Path):
        return str(value)
    return value


def nan2str(value):
    if value is np.nan:
        return ""
    return value


def str2none(value):
    if value == "":
        return None
    return value


def split_name_suffixes(name: str) -> tuple[str, str]:
    """
    Split the base name from its suffixes assuming they are separated by periods.
    """
    # Split only once from the right to get all suffixes correctly
    if "." in name:
        base, suffixes = name.split(".", maxsplit=1)
        suffixes = f".{suffixes}"
    else:
        base = name
        suffixes = ""

    return base, suffixes


def find_unique_name(name: str, names: list[str], case_sensitive=True) -> str:
    """
    Generate a unique name not in `names`.
    If the name ends with (n), increment n until unique.
    For files with extensions, insert the counter before all extensions.

    :param name: Proposed name.
    :param names: List of names to avoid.
    :return: A unique name.
    """
    if not case_sensitive:
        names_list = [val.lower() if isinstance(val, str) else val for val in names]
    else:
        names_list = names

    checkname = name.lower() if not case_sensitive and isinstance(name, str) else name
    if checkname not in names_list:
        return name

    base_part, suffixes = split_name_suffixes(name)

    # Extract and increment count if '(n)' is present
    match = re.match(r"^(.*?)(?:\((\d+)\))?$", base_part)
    base = match.group(1)  # type: ignore
    count = int(match.group(2)) + 1 if match.group(2) else 1  # type: ignore

    while True:
        candidate = f"{base}({count}){suffixes}"
        checkname = candidate if case_sensitive else candidate.lower()

        if checkname not in names_list:
            return candidate

        count += 1


def remove_duplicates_in_list(input_list: list) -> list:
    """
    Remove duplicates from a list without changing the sorting.

    :param input_list: the list to remove duplicates from.

    :return: The sorted list
    """
    return sorted(set(input_list), key=input_list.index)


def decode_byte_array(values: np.ndarray, data_type: type) -> np.array:
    """
    Decode a byte array to an array of a given data type.

    :param values: The byte array to decode.
    :param data_type: The data type to convert the values to.

    :return: The decoded array.
    """
    return (
        np.char.decode(values, "utf-8") if values.dtype.kind == "S" else values
    ).astype(data_type)


def min_max_scaler(
    values: np.ndarray,
    min_scaler: float = 0.0,
    max_scaler: float = 1.0,
    axis: None | int = None,
) -> np.ndarray:
    """
    Min-Max scale an array.

    :param values: The array to scale.
    :param min_scaler: The minimum value to scale to.
    :param max_scaler: The maximum value to scale to.
    :param axis: Axis to apply scaling (eg. 0 for columns, 1 for rows).

    :return: The scaled array.
    """
    # replace NaN with min_scaler
    values = np.nan_to_num(values, nan=min_scaler)

    v_min = values.min(axis=axis, keepdims=True)
    v_max = values.max(axis=axis, keepdims=True)
    v_range = v_max - v_min

    scaled = np.where(
        v_range == 0,
        min_scaler,
        (values - v_min) / v_range * (max_scaler - min_scaler) + min_scaler,
    )

    return scaled


def array_is_colour(values: np.ndarray) -> bool:
    """
    Check if the values are RGB or RGBA.
    The function does not consider the type as we are formatting it.

    :param values: The values to check.

    :return: True if the values are RGB or RGBA.
    """
    if not isinstance(values, np.ndarray):
        return False

    if values.dtype.names:
        if values.dtype.names in (("r", "g", "b"), ("r", "g", "b", "a")) and all(
            np.issubdtype(values.dtype[name], np.number) for name in values.dtype.names
        ):
            return True
        return False

    return (
        values.ndim == 2
        and values.shape[1] in (3, 4)
        and np.issubdtype(values.dtype, np.number)
    )


def format_numeric_values(
    input_values: np.ndarray, n_decimals: int, max_chars: int
) -> np.ndarray:
    """
    Format numeric values for display.

    For values that are too long, scientific notation is used.
    If the value is less than 1, it is rounded to a number of decimals depending on its magnitude.
    Trailing zeros and decimal points are removed.

    :param input_values: The array of values to format.
    :param n_decimals: The number of decimal places to round to.
    :param max_chars: The maximum number of characters for each formatted value.

    :return: An array of formatted strings.
    """
    nan_mask = np.isnan(input_values)
    values = input_values[~nan_mask]

    # prepare "normal format" strings
    mask = (np.abs(values) < 1) & (values != 0)
    decimals = np.full(values.shape, n_decimals, dtype=int)
    if np.any(mask):
        log_abs = np.log10(np.abs(values[mask]))
        decimals[mask] = n_decimals - np.floor(log_abs).astype(int) - 1

    formats = np.char.add(np.char.add("%.", decimals.astype(str)), "f")
    normal_str = np.char.rstrip(np.char.mod(formats, values), "0")

    # prepare the scientific notation strings
    sci_str = np.char.mod(f"%.{n_decimals}e", values)
    man_exp = np.array(np.char.split(sci_str, "e").tolist())
    man = np.char.rstrip(np.char.rstrip(man_exp[:, 0], "0"), ".")
    sci_str = np.char.add(np.char.add(man, "e"), man_exp[:, 1])

    # choose between normal and scientific notation
    result = np.where(np.char.str_len(normal_str) > max_chars, sci_str, normal_str)

    # replace NaN values with empty strings
    final_result = np.full(input_values.shape, "", dtype=object)
    final_result[~nan_mask] = result

    return final_result


def get_unique_name_from_entities(
    name: str,
    entities: list[Any],
    key: str = "name",
    types: type | tuple[type] | None = None,
) -> str:
    """
    Find a unique name in an object, optionally filtering by type.

    :param name: Proposed name.
    :param entities: The list of entities to search in.
    :param key: The key of the object to extract
    :param types: If provided, only entities of this type will be considered.

    :return: A unique name.
    """

    def child_check(child_: Any) -> bool:
        """
        Function to filter entities based on instance type.
        """
        if types is not None and not isinstance(child_, types):
            return False
        return True

    names = []
    for child in entities:
        if child_check(child):
            sub_name = getattr(child, key, None)
            if isinstance(sub_name, str):
                names.append(sub_name)

    return find_unique_name(name, names)


def extract_uids(values) -> list[UUID] | None:
    """
    Extract the UUIDs from a list of UUIDs, Data objects or strings.

    :param values: A list of UUIDs, Data objects or strings.

    :return: A list of UUIDs or None if input is None.
    """
    if values is None:
        return None

    if not isinstance(values, Iterable) or isinstance(values, (str, bytes)):
        values = [values]

    uids = []
    for child in values:
        uid = entity2uuid(str2uuid(child))
        if not isinstance(uid, UUID):
            raise TypeError(
                f"'{child}' must be of type UUID, "
                f"or has 'uid' as attribute, not '{type(child)}'"
            )
        uids.append(uid)

    return uids


def copy_dict_relatives(
    values: dict, parent: EntityContainer | Workspace, clear_cache: bool = False
):
    """
    Copy the objects and groups referenced in a dictionary of values to a new parent.

    The input dictionary is not modified. The values must be already promoted.

    :param values: A dictionary of values possibly containing references to objects and groups.
    :param parent: The parent to copy the objects and groups to.
    :param clear_cache: If True, clear the array attributes of the copied objects and groups.
    """

    # 2. do the copy
    def copy_obj_and_group(val: Any) -> Any:
        """
        Function to copy objects and groups found in the options.
        To be used in dict_mapper for intricate structures.

        :param val: The value to check and possibly copy.

        :return: The same value
        """
        if hasattr(val, "children"):
            if val.workspace.h5file == parent.workspace.h5file:
                raise ValueError("Cannot copy objects within the same workspace.")

            # do not copy if the uuid already exists in the parent workspace
            if parent.workspace.get_entity(val.uid)[0] is not None:
                return val

            val.copy(parent, copy_children=True, clear_cache=clear_cache)  # type: ignore

        return val

    dict_mapper(values, [copy_obj_and_group])


def workspace2path(value):
    if hasattr(value, "h5file"):
        if isinstance(value.h5file, BytesIO):
            return "[in-memory]"
        return str(value.h5file)
    return value


def dict_to_json_str(data: dict) -> str:
    """
    Format all values in a dictionary for json serialization.

    :param data: Dictionary of values to be converted.

    :return: A json string representation of the dictionary.
    """
    formatted = stringify(data)
    formatted = dict_mapper(
        formatted, [lambda x: f"{x:.4e}" if isinstance(x, float) else x]
    )
    return dumps(formatted, indent=4)


def uuid_from_values(data: dict | str) -> UUID:
    """
    Create a deterministic uuid of a dictionary or its json string representation.

    Floats are formatted to fixed precision scientific notation and objects are
    converted to uid strings.

    :param data: Dictionary or a string representation of a dictionary containing
    parameters/values of an application.

    :returns: Unique but recoverable uuid file identifier string.
    """
    if isinstance(data, dict):
        data = dict_to_json_str(data)

    return uuid5(NAMESPACE_DNS, str(data))



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import warnings
from io import BytesIO
from logging import getLogger
from pathlib import Path
from shutil import copy, move
from time import time
from typing import Any

from geoh5py import Workspace
from geoh5py.groups import ContainerGroup, Group
from geoh5py.objects import ObjectBase
from geoh5py.shared.utils import fetch_active_workspace


logger = getLogger(__name__)


def flatten(ui_json: dict[str, dict[str, Any]]) -> dict[str, Any]:
    """
    Flatten ui.json format to simple key/value pairs.

    Converts the nested ui.json structure to a flat dictionary where most fields
    have their values extracted from the 'value' member. For widgets storing
    multiple values, the values are passed as a dictionary.

    :param ui_json: The form containing the UI data to flatten.

    :return: A flattened dictionary with parameter names as keys and their values.
    """
    data: dict[str, Any] = {}
    for name, value in ui_json.items():
        if isinstance(value, dict):
            if is_form(value):
                field = "value" if truth(ui_json, name, "isValue") else "property"
                if not truth(ui_json, name, "enabled"):
                    data[name] = None
                elif (
                    "groupValue" in value
                    and "value" in value
                    and value["groupValue"] is not None
                ):
                    data[name] = {
                        "groupValue": value["groupValue"],
                        "value": value["value"],
                    }
                elif (
                    "isComplement" in value
                    and "property" in value
                    and value["property"] is not None
                ):
                    data[name] = {
                        "value": value["value"],
                        "isComplement": value["isComplement"],
                        "property": value["property"],
                    }
                else:
                    data[name] = value[field]

        else:
            data[name] = value

    return data


def collect(
    ui_json: dict[str, dict[str, Any]], member: str, value: Any = None
) -> dict[str, dict[str, Any]]:
    """
    Collect UI parameters with a common field and optional value.

    Searches through the ui_json dictionary for parameters that contain a
    specific member field, optionally filtering by the value of that field.

    :param ui_json: The UI JSON dictionary to search through.
    :param member: The field name to search for in each parameter.
    :param value: Optional value to match against the member field.

    :return: Dictionary of parameters that match the criteria.
    """

    parameters = {}
    for name, form in ui_json.items():
        if is_form(form) and member in form:
            if value is None or form[member] == value:
                parameters[name] = form

    return parameters


def find_all(
    ui_json: dict[str, dict[str, Any]], member: str, value: Any = None
) -> list[str]:
    """
    Return names of all parameters matching the given criteria.

    Convenience function that returns just the parameter names from the collect
    function results.

    :param ui_json: The UI JSON dictionary to search through.
    :param member: The field name to search for in each parameter.
    :param value: Optional value to match against the member field.

    :return: List of parameter names that match the criteria.
    """
    parameters = collect(ui_json, member, value)
    return list(parameters.keys())


def group_optional(ui_json: dict[str, dict[str, Any]], group_name: str) -> bool:
    """
    Check if a group has the groupOptional flag enabled.

    Searches for parameters belonging to the specified group and returns the
    groupOptional boolean value for that group.

    :param ui_json: The UI JSON dictionary to search through.
    :param group_name: Name of the group to check.

    :return: True if the group is optional, False otherwise.
    """
    group = collect(ui_json, "group", group_name)
    parameters = find_all(group, "groupOptional")
    return group[parameters[0]]["groupOptional"] if parameters else False


def optional_requires_value(ui_json: dict[str, dict[str, Any]], parameter: str) -> bool:
    """
    Check if an optional parameter requires a value based on its enabled state.

    For optional parameters, this function checks the 'enabled' field to determine
    if the parameter requires a value.

    :param ui_json: UI JSON dictionary containing the parameters.
    :param parameter: Name of the parameter to check.

    :return: True if the parameter is enabled, False otherwise.
    """
    return ui_json[parameter].get("enabled", True)


def dependency_requires_value(
    ui_json: dict[str, dict[str, Any]], parameter: str
) -> bool:
    """
    Check if a parameter requires a value based on its dependency constraints.

    Handles dependency and optional requirements for parameters. If a dependency
    doesn't require a value then the function returns False. If the dependency
    does require a value, the return value is either True, or will take on the
    enabled state if the dependent parameter is optional.

    :param ui_json: UI JSON dictionary containing the parameters.
    :param parameter: Name of the parameter to check.

    :return: True if the parameter requires a value based on dependencies, False otherwise.
    """
    dependency = ui_json[parameter]["dependency"]

    key = (
        "enabled"
        if (
            ui_json[dependency].get("optional", False)
            or "enabled" in ui_json[dependency]
        )
        else "value"
    )

    if ui_json[parameter].get("dependencyType", "enabled") == "enabled":
        is_required = bool(ui_json[dependency].get(key, True)) & bool(
            ui_json[dependency]["value"]
        )
    else:
        is_required = not ui_json[dependency].get(key, True)
    if ("optional" in ui_json[parameter]) & bool(is_required):
        is_required = ui_json[parameter]["enabled"]

    return is_required


def group_requires_value(ui_json: dict[str, dict[str, Any]], parameter: str) -> bool:
    """
    Check if a parameter requires a value based on its group's optional status.

    Determines if a parameter requires a value by checking if its group is optional
    and if the group is enabled.

    :param ui_json: UI JSON dictionary containing the parameters.
    :param parameter: Name of the parameter to check.

    :return: True if the group is optional and enabled, or if the group is not optional.
    """
    is_required = True
    groupname = ui_json[parameter]["group"]
    group = collect(ui_json, "group", groupname)
    if group_optional(ui_json, groupname):
        is_required = group_enabled(group)
    return is_required


def requires_value(ui_json: dict[str, dict[str, Any]], parameter: str) -> bool:
    """
    Check if a UI JSON parameter requires a value based on hierarchy of switches.

    The required status of a parameter depends on a hierarchy of UI switches.
    At the top is the groupOptional switch, below that is the dependency
    switch, and at the bottom is the optional switch. When groupOptional
    is disabled, all parameters in the group are not required. When the
    groupOptional is enabled, the required status of a parameter depends first
    on any dependencies and lastly on its optional status.

    :param ui_json: UI JSON dictionary containing the parameters.
    :param parameter: Name of the parameter to check.

    :return: True if the parameter requires a value, False otherwise.
    """
    is_required = True

    if is_form(ui_json[parameter]):
        if "group" in ui_json[parameter]:
            group_required = group_requires_value(ui_json, parameter)
            if group_required:
                if "dependency" in ui_json[parameter]:
                    is_required = dependency_requires_value(ui_json, parameter)
                elif "optional" in ui_json[parameter]:
                    is_required = optional_requires_value(ui_json, parameter)
            else:
                is_required = False

        elif "dependency" in ui_json[parameter]:
            is_required = dependency_requires_value(ui_json, parameter)

        elif "optional" in ui_json[parameter]:
            is_required = optional_requires_value(ui_json, parameter)

    return is_required


def group_enabled(group: dict[str, dict[str, Any]]) -> bool:
    """
    Check if a group is enabled based on groupOptional and enabled flags.

    Returns True if both groupOptional and enabled are True for the group.

    :param group: UI JSON group dictionary containing parameters.

    :return: True if the group is enabled, False otherwise.

    :raises ValueError: If the provided group does not contain a
        parameter with groupOptional member.
    """
    parameters = find_all(group, "groupOptional")
    if not parameters:
        raise ValueError(
            "Provided group does not contain a parameter with groupOptional member."
        )
    return group[parameters[0]].get("enabled", True)


def set_enabled(
    ui_json: dict[str, Any], parameter: str, value: bool, validate: bool = True
) -> None:
    """
    Set the enabled status for an optional or groupOptional parameter.

    Updates the 'enabled' field for parameters that support it, including optional
    parameters, parameters with dependencies, and group optional parameters.

    :param ui_json: UI JSON dictionary containing the parameters.
    :param parameter: Name of the parameter to modify.
    :param value: Boolean value to set for the parameter's enabled member.
    :param validate: Whether to perform validation checks on the operation.

    :raises Warning: If attempting to disable a non-optional parameter.
    """
    if ui_json[parameter].get("optional", False) or bool(
        ui_json[parameter].get("dependency", False)
    ):
        ui_json[parameter]["enabled"] = value

    is_group_optional = False
    group_name = ui_json[parameter].get("group", False)
    if group_name:
        group = collect(ui_json, "group", group_name)
        parameters = find_all(group, "groupOptional")
        if parameters:
            is_group_optional = True
            if parameters[0] == parameter:
                for form in group.values():
                    form["enabled"] = value

    if validate:
        if not is_group_optional and "dependency" in ui_json[parameter]:
            is_group_optional = not dependency_requires_value(ui_json, parameter)

        if (not value) and not (
            ui_json[parameter].get("optional", False) or is_group_optional
        ):
            warnings.warn(
                f"Non-option parameter '{parameter}' cannot be set to 'enabled' False "
            )


def truth(ui_json: dict[str, dict[str, Any]], name: str, member: str) -> bool:
    """
    Get a parameter's member value with default fallback for non-existent members.

    Returns the value of a specific member field for a parameter, or a default
    value if the member doesn't exist.

    :param ui_json: UI JSON dictionary containing the parameters.
    :param name: Name of the parameter to check.
    :param member: Name of the member field to retrieve.

    :return: The member's value or its default value.

    :raises ValueError: If the field was not provided and does not have a default state.
    """
    default_states = {
        "enabled": True,
        "optional": False,
        "groupOptional": False,
        "main": False,
        "isValue": True,
    }
    if member in ui_json[name]:
        return ui_json[name][member]

    if member in default_states:
        return default_states[member]

    raise ValueError(
        f"Field: {member} was not provided in ui.json and does not have a default state."
    )


def is_uijson(ui_json: dict[str, dict[str, Any]]) -> bool:
    """
    Check if a dictionary contains all the required UI JSON parameters.

    Validates that a dictionary has the structure and required fields of a
    valid UI JSON configuration.

    :param ui_json: Dictionary to validate as UI JSON.

    :return: True if the dictionary contains all required parameters, False otherwise.
    """
    required_parameters = [
        "title",
        "monitoring_directory",
        "run_command",
        "conda_environment",
        "geoh5",
        "workspace_geoh5",
    ]

    is_a_uijson = True
    for k in required_parameters:
        if k not in ui_json.keys():
            is_a_uijson = False

    return is_a_uijson


def is_form(var: Any) -> bool:
    """
    Check if a variable is a valid form dictionary.

    Returns True if the dictionary variable contains both 'label' and 'value' members,
    which are required for a valid UI form element.

    :param var: Variable to check for form structure.

    :return: True if the variable is a valid form dictionary, False otherwise.
    """
    is_a_form = False
    if isinstance(var, dict):
        if all(k in var.keys() for k in ["label", "value"]):
            is_a_form = True

    return is_a_form


def str2list(value: Any) -> Any:
    """
    Convert string representation of numbers to list of floats.

    Maps string values like "[1,2,3]" or "1,2,3" to [1.0, 2.0, 3.0].
    Handles special cases like "inf", "-inf", and empty strings.

    :param value: Value to convert, typically a string or already converted type.

    :return: List of floats if conversion successful, original value otherwise.
    """
    if isinstance(value, str):
        if value in ["inf", "-inf", ""]:
            return value
        try:
            return [float(n) for n in value.split(",") if n != ""]
        except ValueError:
            return value

    return value


def str2inf(value: Any) -> Any:
    """
    Convert string representations of infinity to float infinity values.

    Converts "inf" and "-inf" strings to their corresponding float values.

    :param value: Value to convert, typically a string.

    :return: Float infinity value if input is "inf" or "-inf", original value otherwise.
    """
    if value in ["inf", "-inf"]:
        return float(value)
    return value


def workspace2path(value: Any) -> str:
    """
    Convert a Workspace object to its file path string.

    **Deprecated:** This function has been migrated to `geoh5py.shared.utils.workspace2path`
    and will be removed in future versions.

    :param value: Workspace object or other value to convert.

    :return: File path string for Workspace objects, "[in-memory]" for in-memory workspaces,
             or the original value for non-Workspace inputs.
    """
    logger.warning(
        "Deprecation Warning - This function has been migrated to "
        "`geoh5py.shared.utils.workspace2path` and will be removed in"
        "future versions.",
    )
    if isinstance(value, Workspace):
        if isinstance(value.h5file, BytesIO):
            return "[in-memory]"
        return str(value.h5file)
    return value


def path2workspace(value: Any) -> Any:
    """
    Convert a file path string to a Workspace object if it represents a valid geoh5 file.

    Creates a Workspace object from a path string if the file exists and has a .geoh5 extension.

    :param value: File path string, Path object, or other value to convert.

    :return: Workspace object if the value represents a valid geoh5 file, original value otherwise.
    """
    if (
        isinstance(value, (str, Path))
        and Path(value).suffix == ".geoh5"
        and Path(value).exists()
    ):
        workspace = Workspace(value, mode="r")
        workspace.close()
        return workspace
    return value


def container_group2name(value: Any) -> Any:
    """
    Convert a ContainerGroup object to its name string.

    **Deprecated:** This function will be removed in future releases.

    :param value: ContainerGroup object or other value to convert.

    :return: Name string for ContainerGroup objects, original value otherwise.
    """
    logger.warning(
        "Deprecation Warning - This function will be removed in future releases."
    )
    if isinstance(value, ContainerGroup):
        return value.name
    return value


def monitored_directory_copy(
    directory: str, entity: ObjectBase | Group, copy_children: bool = True
) -> str:
    """
    Create a temporary geoh5 file in the monitoring folder and export entity for update.

    Creates a temporary workspace in the specified monitoring directory and copies
    the given entity to it. This is useful for monitoring and updating entities
    in a separate workspace environment.

    :param directory: Path to the monitoring directory where the temporary file will be created.
    :param entity: Entity (ObjectBase or Group) to be copied for monitoring.
    :param copy_children: Whether to copy children entities along with the main entity.

    :return: Full path to the created temporary geoh5 file.
    """
    directory_path = Path(directory)
    working_path = directory_path / ".working"
    working_path.mkdir(exist_ok=True)

    temp_geoh5 = f"temp{time():.3f}.geoh5"

    with fetch_active_workspace(entity.workspace, mode="r"):
        with Workspace.create(working_path / temp_geoh5) as w_s:
            entity.copy(parent=w_s, copy_children=copy_children)

    move(working_path / temp_geoh5, directory_path / temp_geoh5, copy)

    return str(directory_path / temp_geoh5)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

# pylint: disable=too-many-instance-attributes, too-many-arguments

from __future__ import annotations

import uuid
from typing import cast

import numpy as np

from geoh5py.data import Data, FloatData, IntegerData, PrimitiveTypeEnum, ReferencedData
from geoh5py.objects import DrapeModel, Grid2D
from geoh5py.objects.grid_object import GridObject
from geoh5py.shared.utils import KEY_MAP, str2uuid, xy_rotation_matrix


class VPModel(GridObject, DrapeModel):
    """
    VP Mesh object defined by a uniform grid in the u and v directions, and
    draped layers and prisms that represent the vertical structure of the model.

    Sub-class of :obj:`~geoh5py.objects.grid_object.GridObject` and
    :obj:`~geoh5py.objects.drape_model.DrapeModel`.

    :param u_cell_size: Cell size along the u-axis.
    :param u_count: Number of cells along the u-axis.
    :param v_cell_size: Cell size along the v-axis.
    :param v_count: Number of cells along the v-axis.
    :param flag_property_id: UUID or name of the flag property.
    :param heterogeneous_property_id: UUID or name of the heterogeneous property.
    :param physical_data_name: Name of the physical data.
    :param unit_property_id: UUID or name of the unit property.
    :param weight_property_id: UUID or name of the weight property.
    """

    _VALUE_MAP = {
        0: "Unknown",
        100000: "VP_basement",
    }
    _TYPE_UID = uuid.UUID("{7d37f28f-f379-4006-984e-043db439ee95}")
    _LAYERS_DTYPE = np.dtype([("I", "<i4"), ("J", "<i4"), ("Bottom elevation", "<f8")])
    _PRISM_DTYPE = np.dtype(
        [
            ("Top elevation", "<f8"),
            ("First layer", "<i4"),
            ("Layer count", "<i4"),
        ]
    )
    _attribute_map = GridObject._attribute_map.copy()
    _attribute_map.update(
        {
            "Nu": "u_count",
            "Nv": "v_count",
            "Origin": "origin",
            "Rotation": "rotation",
            "U size": "u_cell_size",
            "V size": "v_cell_size",
            "Flag property ID": "flag_property_id",
            "Heterogeneous property ID": "heterogeneous_property_id",
            "Physical data name": "physical_data_name",
            "Unit property ID": "unit_property_id",
            "Weight property ID": "weight_property_id",
        }
    )

    def __init__(
        self,
        u_cell_size: float = 1.0,
        u_count: int = 1,
        v_cell_size: float = 1.0,
        v_count: int = 1,
        *,
        flag_property_id: str | uuid.UUID | np.ndarray = np.asarray([]),
        heterogeneous_property_id: str | uuid.UUID | np.ndarray = np.asarray([]),
        physical_data_name: str | np.ndarray = "Property",
        unit_property_id: str | uuid.UUID | np.ndarray = np.asarray([]),
        weight_property_id: str | uuid.UUID | np.ndarray = np.asarray([]),
        **kwargs,
    ):
        self._u_count: np.int32 = self.validate_count(u_count, "u")
        self._v_count: np.int32 = self.validate_count(v_count, "v")

        super().__init__(**kwargs)

        self.u_cell_size: float = u_cell_size
        self.v_cell_size: float = v_cell_size
        self.unit_property_id = self._promote_uuid_attribute(
            unit_property_id,
            "unit_property_id",
            PrimitiveTypeEnum.REFERENCED,
        )
        self.flag_property_id = self._promote_uuid_attribute(
            flag_property_id, "flag_property_id", PrimitiveTypeEnum.INTEGER
        )
        self.heterogeneous_property_id = self._promote_uuid_attribute(
            heterogeneous_property_id,
            "heterogeneous_property_id",
            PrimitiveTypeEnum.FLOAT,
        )
        self.physical_data_name = physical_data_name
        self.weight_property_id = self._promote_uuid_attribute(
            weight_property_id, "weight_property_id", PrimitiveTypeEnum.FLOAT
        )

    def _promote_uuid_attribute(
        self,
        value: str | uuid.UUID | np.ndarray | Data,
        name: str,
        primitive_type: PrimitiveTypeEnum,
    ) -> uuid.UUID:
        """
        Promote a string or UUID to a UUID.

        :param value: The value provided in the constructor.
        :param name: The name of the attribute validated.
        :param primitive_type: The primitive type of the attribute.
        """
        if isinstance(value, Data):
            if value not in self.children:
                raise ValueError(
                    f"Data '{value.name}' is not a child of the VPModel object."
                )
            return value.uid

        value = str2uuid(value)

        if isinstance(value, uuid.UUID):
            return value

        if not isinstance(value, np.ndarray):
            raise TypeError(
                f"Attribute '{name}' should be a 'uuid.UUID' or an array of values."
            )

        kwargs = {"association": "CELL", "type": primitive_type, "values": value}
        data = cast(
            Data,
            self.add_data(
                {KEY_MAP[name]: kwargs},
            ),
        )

        if isinstance(data, ReferencedData) and data.value_map is not None:
            value_map = data.value_map()
            value_map.update(self._VALUE_MAP)
            data.entity_type.value_map = data.entity_type.validate_value_map(value_map)

        return data.uid

    @classmethod
    def create(cls, workspace, **kwargs):
        """
        Function to create an entity.

        The visual parameters are set to default values, and the filter_basement
        is set to 5% of the vertical extent of the model.

        :param workspace: Workspace to be added to.
        :param kwargs: List of keyword arguments defining the properties of a class.

        :return entity: Registered Entity to the workspace.
        """
        new_object = super().create(workspace, **kwargs)

        viz_params = new_object.add_default_visual_parameters()
        viz_params.filter_basement = (
            new_object.prisms[:, 0].max() - new_object.layers[:, 2].min()
        ) * 0.05

        return new_object

    @property
    def flag_property_id(self) -> uuid.UUID:
        """
        VPmg flag property.
        """
        return self._flag_property_id

    @flag_property_id.setter
    def flag_property_id(self, value: uuid.UUID | IntegerData):
        if isinstance(value, IntegerData):
            value = value.uid

        if not isinstance(value, uuid.UUID):
            raise TypeError("Attribute 'flag_property_id' should be a 'uuid.UUID'.")

        # TODO: Check that the UUID belong to a child of the VPModel.
        # Requires to change the loading mechanisms such that data
        # are loaded at the same time as the object.
        self._flag_property_id = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def heterogeneous_property_id(self) -> uuid.UUID:
        """
        VPmg heterogenenous model id.
        """
        return self._heterogeneous_property_id

    @heterogeneous_property_id.setter
    def heterogeneous_property_id(self, value: uuid.UUID | FloatData):
        if isinstance(value, FloatData):
            value = value.uid

        if not isinstance(value, uuid.UUID):
            raise TypeError(
                "Attribute 'heterogeneous_property_id' should be a 'uuid.UUID'."
            )

        # TODO: Check that the UUID belong to a child of the VPModel.
        # Requires to change the loading mechanisms such that data
        # are loaded at the same time as the object.
        self._heterogeneous_property_id = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def physical_data_name(self) -> str:
        """
        VPmg physical property model data map name.
        """
        return self._physical_data_name

    @physical_data_name.setter
    def physical_data_name(self, value: str):
        if not isinstance(value, str):
            raise TypeError("Attribute 'physical_data_name' should be a 'str'")

        self._physical_data_name = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def unit_property_id(self) -> uuid.UUID:
        """
        VPmg unit model id.
        """
        return self._unit_property_id

    @unit_property_id.setter
    def unit_property_id(self, value: uuid.UUID | ReferencedData):
        if isinstance(value, ReferencedData):
            value = value.uid

        if not isinstance(value, uuid.UUID):
            raise TypeError("Attribute 'unit_property_id' should be a 'uuid.UUID'.")

        # TODO: Check that the UUID belong to a child of the VPModel.
        # Requires to change the loading mechanisms such that data
        # are loaded at the same time as the object.
        self._unit_property_id = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def weight_property_id(self) -> uuid.UUID:
        """
        VPmg weight model id.
        """
        return self._weight_property_id

    @weight_property_id.setter
    def weight_property_id(self, value: uuid.UUID | FloatData):
        if isinstance(value, FloatData):
            value = value.uid

        if not isinstance(value, uuid.UUID):
            raise TypeError("Attribute 'weight_property_id' should be a 'uuid.UUID'.")

        # TODO: Check that the UUID belong to a child of the VPModel.
        # Requires to change the loading mechanisms such that data
        # are loaded at the same time as the object.
        self._weight_property_id = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def centroids(self) -> np.ndarray:
        """
        Cell center locations of each prism in world coordinates, shape(*, 3).

        .. code-block:: python

            centroids = [
                [x_1, y_1, z_1],
                ...,
                [x_N, y_N, z_N]
            ]

        The positions are computed from the u, v cell sizes and rotation angle
        to determine the horizontal position. The vertical position is determined
        by the mean of elevation of the top and bottom of the prism.
        """
        if getattr(self, "_centroids", None) is None:
            # Create a grid of u, v coordinates based on the cell sizes
            rotation_matrix = xy_rotation_matrix(np.deg2rad(self.rotation))
            v_grid, u_grid = np.meshgrid(
                np.cumsum(np.ones(self._v_count) * self.v_cell_size)
                - self.v_cell_size / 2,
                np.cumsum(np.ones(self._u_count) * self.u_cell_size)
                - self.u_cell_size / 2,
            )
            xyz = (
                rotation_matrix
                @ np.c_[np.ravel(u_grid), np.ravel(v_grid), self.prisms[:, 0]].T
            ).T
            xyz += np.asarray(self._origin.tolist())[None, :]

            # Instantiate an array of centroids, one for each prism
            indices = (self.layers[:, 0] * self._v_count + self.layers[:, 1]).astype(
                int
            )
            centroids = xyz[indices, :]

            # Loop down the layers and compute the center, up to the deepest layer
            # Values for centroids are only updated if the layer is deeper than
            # the previous one. Elevation is the top of the prism of the nth layer.
            elevation = (
                self.prisms[:, 0].reshape((self.v_count, self.u_count)).T.flatten()
            )
            ind_colm = (
                self.prisms[:, 2].reshape((self.v_count, self.u_count)).T.flatten()
            )
            top_ind = np.r_[0, np.cumsum(ind_colm[:-1])].astype(int)
            for count in range(int(ind_colm.max()) - 1):
                mask = ind_colm > (count + 1)
                centroids[top_ind, 2] = np.mean(
                    [elevation, self.layers[top_ind, 2]], axis=0
                )
                elevation[mask] = self.layers[top_ind[mask], 2]
                top_ind[mask] += 1

            self._centroids = centroids

        return self._centroids

    @property
    def n_cells(self) -> int:
        """
        Total number of cells
        """
        return int(self.shape[2])

    @property
    def shape(self) -> np.ndarray:
        """
        Shape of the drape model in terms of number of prisms and layers.
        """
        return np.array([self._u_count, self._v_count, self.layers.shape[0]])

    @property
    def u_cell_size(self) -> float:
        """
        Cell size along the u-axis.
        """
        return self._u_cell_size

    @u_cell_size.setter
    def u_cell_size(self, value: float):
        self._u_cell_size = Grid2D.validate_cell_size(value, "u")
        self._centroids = None

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def u_count(self) -> np.int32:
        """
        Number of cells along the u-axis.
        """
        return self._u_count

    @property
    def v_cell_size(self) -> float:
        """
        Cell size along the v-axis.
        """
        return self._v_cell_size

    @v_cell_size.setter
    def v_cell_size(self, value: float):
        self._v_cell_size = Grid2D.validate_cell_size(value, "v")
        self._centroids = None

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def v_count(self) -> np.int32:
        """
        Number of cells along the v-axis.
        """
        return self._v_count

    @staticmethod
    def validate_count(value: int, axis: str) -> np.int32:
        """
        Validate and format type of count value.
        """
        if not isinstance(value, (np.integer, int)) or value < 1:
            raise TypeError(
                f"Attribute '{axis}_count' must be a type(int32) greater than 0."
            )

        return np.int32(value)

    def validate_metadata(self, value: np.ndarray | dict | None) -> dict | None:
        """
        Validate and format type of metadata value.

        :param value: Dictionary of values to be augmented with the "VP" field.
        """

        value = super().validate_metadata(value)

        if value is None:
            value = {}

        if value.get("VP", None) is None:
            value["VP"] = {"Base of model elevation (m)": self.layers[:, 2].min()}

        return value



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

import json
from typing import Any
from uuid import UUID

import numpy as np

from geoh5py.shared.utils import (
    copy_dict_relatives,
    dict_mapper,
    entity2uuid,
    str2uuid,
    str_json_to_dict,
    stringify,
)

from .base import Group


class UIJsonGroup(Group):
    """
    Group for storing ui.json files.

    In Geoscience ANALYST, it can be used to store ui.jsons.

    :param options: Dictionary containing the ui.json structure.
    """

    _TYPE_UID = UUID("{BB50AC61-A657-4926-9C82-067658E246A0}")
    _default_name = "UIJson"

    def __init__(
        self,
        options: dict | np.ndarray | bytes | None = None,
        **kwargs,
    ):
        self._options: dict

        super().__init__(**kwargs)

        self.options = self.format_input_options(options)

    def _copy_relatives(self, parent, clear_cache: bool = False):
        """
        Copy the relatives of the entity.

        :param parent: The parent to copy the entity to.
        :param clear_cache: Indicate whether to clear the cache.
        """

        if (
            parent is None
            or parent == self.parent
            or parent == self.workspace
            or len(self.options) == 0
        ):
            return

        options = self.options.copy()
        options.pop("out_group", None)

        copy_dict_relatives(
            self.workspace.promote(options), parent, clear_cache=clear_cache
        )

    def _prepare_options(self, options: dict) -> dict:
        """
        Prepare the out_group entry in the options dictionary.

        It added the geoh5 file path, and an out_group entry.
        Templates cannot be used due to circular import.

        :param options: The options dictionary to prepare.

        :return: The prepared options dictionary.
        """
        options["geoh5"] = str(self.workspace.h5file)

        out_group = options.get("out_group", {}) or {}
        out_group["value"] = str(self.uid)
        out_group["label"] = out_group.get("label", self.name)
        out_group["groupType"] = str(self.default_type_uid())
        options["out_group"] = out_group

        return options

    def copy(
        self,
        parent=None,
        *,
        copy_children: bool = False,
        copy_relatives: bool = True,
        clear_cache: bool = False,
        **kwargs,
    ) -> UIJsonGroup | None:
        """
        Sub-class extension of :func:`~geoh5py.groups.base.Group.copy`.

        :param parent: The parent to copy the entity to.
        :param copy_children: Whether to copy the children of the entity.
        :param copy_relatives: If true, the objects and groups referenced in the options are copied.
        :param clear_cache: Indicate whether to clear the cache.
        :param kwargs: other keyword arguments.

        :return: The copied entity.
        """
        if copy_relatives:
            self._copy_relatives(parent, clear_cache=clear_cache)

        copied = super().copy(
            parent=parent,
            copy_children=copy_children,
            clear_cache=clear_cache,
            **kwargs,
        )

        return copied

    @property
    def options(self) -> dict:
        """
        Metadata attached to the entity.

        Return a copy of the dictionary to avoid accidental modifications.
        """
        return self._options.copy()

    @options.setter
    def options(self, value: dict):
        if not isinstance(value, dict):
            raise TypeError(f"Input 'options' must be of type {dict}.")

        self._options = dict_mapper(value, [str2uuid, entity2uuid])

        if len(self._options) > 0:
            self._options = self._prepare_options(self._options)

        if self.on_file:
            self.workspace.update_attribute(self, "options")

    def modify_option(self, key: str, value: Any):
        """
        Modify a single option in the options dictionary.

        :param key: the key to modify
        :param value: The value to set
        """
        if key in ["geoh5", "out_group"]:
            raise ValueError(f"Cannot modify the '{key}' entry of the options.")

        self._options[key] = stringify(value)

        if self.on_file:
            self.workspace.update_attribute(self, "options")

    def add_ui_json(self, name: str | None = None):
        """
        Add ui.json file to entity.

        :param name: Name of the file in the workspace.
        """
        if not self._options:
            raise ValueError("UIJsonGroup must have options set.")

        json_str = json.dumps(stringify(self.options), indent=4)
        bytes_data = json_str.encode("utf-8")

        if name is None:
            name = self.name

        return self.add_file(bytes_data, name=f"{name}.ui.json")

    @staticmethod
    def format_input_options(options: dict | np.ndarray | bytes | None) -> dict:
        """
        Format input options to a dictionary.

        :param options: Input options.

        :return: Formatted options.
        """
        if options is None:
            return {}

        if isinstance(options, np.ndarray):
            options = options[0]

        if isinstance(options, bytes):
            options = str_json_to_dict(options)

        return options



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import Annotated, Any
from uuid import UUID

import numpy as np
from pydantic import (
    BaseModel,
    ConfigDict,
    PlainSerializer,
    field_serializer,
    field_validator,
    model_validator,
)
from pydantic.alias_generators import to_camel
from pydantic.functional_validators import BeforeValidator

from geoh5py.groups import Group
from geoh5py.objects import ObjectBase
from geoh5py.shared.validators import (
    empty_string_to_none,
    to_class,
    to_list,
    to_path,
    to_uuid,
    types_to_string,
    uuid_to_string,
    uuid_to_string_or_numeric,
)


# pylint: disable=too-many-return-statements


class DependencyType(str, Enum):
    ENABLED = "enabled"
    DISABLED = "disabled"
    SHOW = "show"
    HIDE = "hide"


class BaseForm(BaseModel):
    """
    Base class for uijson forms

    :param label: Label for ui element.
    :param value: The parameter's value.
    :param optional: If True, ui element is rendered with a checkbox to
        control the enabled state.
    :param enabled: If False, ui element is rendered grey and value will
        be written to file as None.
    :param main: Controls whether ui element will render in the general
        parameters tab (True) or optional parameters (False).
    :param tooltip: String rendered on hover over ui element.
    :param group: Grouped ui elements will be rendered within a box labelled
        with the group name.
    :param group_optional: If True, ui group is rendered with a checkbox that
        controls the enabled state of all of the groups members
    :param dependency: Name of parameter that controls the enabled or
        visible state of the ui element.
    :param dependency_type: Controls whether the ui element is enabled
        or visible when the dependency is enabled if optional or True
        if a bool type.
    :param group_dependency: Name of parameter that controls the enabled
        or visible state of the ui group.
    :param group_dependency_type: Controls whether the ui group is
        enabled or visible when the group dependency is enabled if
        optional or True if a bool type.
    :param verbose: Sets the level at which Geoscience Analyst will make
        the parameter visible in a ui.json file.  Verbosity level is set
        within Analyst menu.
    """

    model_config = ConfigDict(
        extra="allow",
        frozen=True,
        populate_by_name=True,
        loc_by_alias=True,
        alias_generator=to_camel,
    )

    label: str
    value: Any
    optional: bool = False
    enabled: bool = True
    main: bool = True
    tooltip: str = ""
    group: str = ""
    group_optional: bool = False
    dependency: str = ""
    dependency_type: DependencyType = DependencyType.ENABLED
    group_dependency: str = ""
    group_dependency_type: DependencyType = DependencyType.ENABLED

    @classmethod
    def infer(cls, data: dict[str, Any]) -> type[BaseForm]:
        """
        Infer and return the appropriate form.

        :param data: Dictionary of form data.
        """
        data = {to_camel(k): v for k, v in data.items()}
        if "choiceList" in data:
            if data.get("multiSelect", False):
                return MultiChoiceForm
            return ChoiceForm
        if "originalLabel" in data and "alternateLabel" in data:
            return RadioLabelForm
        if any(k in data for k in ["fileDescription", "fileType"]):
            return FileForm
        if "meshType" in data:
            return ObjectForm
        if "groupType" in data:
            return GroupForm
        if any(
            k in data
            for k in ["parent", "association", "dataType", "isValue", "property"]
        ):
            return DataForm
        if isinstance(data.get("value"), str):
            return StringForm
        if isinstance(data.get("value"), bool):
            return BoolForm
        if isinstance(data.get("value"), int):
            return IntegerForm
        if isinstance(data.get("value"), float):
            return FloatForm

        raise ValueError(f"Could not infer form from data: {data}")

    @property
    def json_string(self):
        return self.model_dump_json(exclude_unset=True, by_alias=True)

    def flatten(self):
        """Returns the data for the form."""
        return self.value

    def validate_data(self, params: dict[str, Any]):
        """Validate the form data."""


class StringForm(BaseForm):
    """
    String valued uijson form.
    """

    value: str = ""


class RadioLabelForm(StringForm):
    """
    Radio button for two-option strings.

    The uijson dialogue will render two radio buttons with label choices.  Any
    form labels within the ui.json containing the string matching the original
    button will be altered to the reflect the new button choice.

    :param original_label: Label for the original value.
    :param alternative_label: Label for the alternative value.
    """

    original_label: str
    alternate_label: str


class BoolForm(BaseForm):
    """
    Boolean valued uijson form.
    """

    value: bool = True


class IntegerForm(BaseForm):
    """
    Integer valued uijson form.
    """

    value: int = 1
    min: float = -np.inf
    max: float = np.inf


class FloatForm(BaseForm):
    """
    Float valued uijson form.
    """

    value: float = 1.0
    min: float = -np.inf
    max: float = np.inf
    precision: int = 2
    line_edit: bool = True


class ChoiceForm(BaseForm):
    """
    Choice list uijson form.
    """

    value: str
    choice_list: list[str]

    @model_validator(mode="after")
    def valid_choice(self):
        if self.value not in self.choice_list:
            raise ValueError(f"Provided value: '{self.value}' is not a valid choice.")

        return self


class MultiChoiceForm(BaseForm):
    """Multi-choice list uijson form."""

    value: list[str]
    choice_list: list[str]
    multi_select: bool = True

    @field_validator("multi_select", mode="before")
    @classmethod
    def only_multi_select(cls, value):
        if not value:
            raise ValueError("MultiChoiceForm must have multi_select: True.")
        return value

    @field_validator("value", mode="before")
    @classmethod
    def to_list(cls, value):
        if not isinstance(value, list):
            value = [value]
        return value

    @model_validator(mode="after")
    def valid_choice(self):
        bad_choices = []
        for val in self.value:
            if val not in self.choice_list:
                bad_choices.append(val)
        if bad_choices:
            raise ValueError(f"Provided value(s): {bad_choices} are not valid choices.")

        return self


PathList = Annotated[
    list[Path],
    BeforeValidator(to_path),
    BeforeValidator(to_list),
]


class FileForm(BaseForm):
    """
    File path uijson form
    """

    value: PathList
    file_description: list[str]
    file_type: list[str]
    file_multi: bool = False
    directory_only: bool = False

    @field_serializer("value", when_used="json")
    def to_string(self, value):
        return ";".join([str(path) for path in value])

    @field_validator("value")
    @classmethod
    def valid_file(cls, value):
        bad_paths = []
        for path in value:
            if not path.exists():
                bad_paths.append(path)
        if any(bad_paths):
            raise ValueError(f"Provided path(s) {bad_paths} does not exist.")
        return value

    @model_validator(mode="after")
    def same_length(self):
        if len(self.file_description) != len(self.file_type):
            raise ValueError("File description and type lists must be the same length.")
        return self

    @model_validator(mode="after")
    def value_file_type(self):
        bad_paths = []
        for path in self.value:
            if not self.directory_only and Path(path).suffix[1:] not in self.file_type:
                bad_paths.append(path)
        if any(bad_paths):
            raise ValueError(f"Provided paths {bad_paths} have invalid extensions.")
        return self

    @model_validator(mode="before")
    @classmethod
    def directory_file_type(cls, data):
        if data.get("directoryOnly", False) and data["fileType"] != ["directory"]:
            raise ValueError(
                "File type must be ['directory'] if directory_only is True."
            )
        if data.get("directoryOnly", False) and data["fileDescription"] != [
            "Directory"
        ]:
            raise ValueError(
                "File description must be ['Directory'] if directory_only is True."
            )
        return data


MeshTypes = Annotated[
    list[type[ObjectBase]],
    BeforeValidator(to_class),
    BeforeValidator(to_uuid),
    BeforeValidator(to_list),
    PlainSerializer(types_to_string, when_used="json"),
]

OptionalUUID = Annotated[
    UUID | None,  # pylint: disable=unsupported-binary-operation
    BeforeValidator(empty_string_to_none),
    PlainSerializer(uuid_to_string),
]


class ObjectForm(BaseForm):
    """
    Geoh5py object uijson form.
    """

    value: OptionalUUID
    mesh_type: MeshTypes


GroupTypes = Annotated[
    list[type[Group]],
    BeforeValidator(to_class),
    BeforeValidator(to_uuid),
    BeforeValidator(to_list),
    PlainSerializer(types_to_string, when_used="json"),
]


class GroupForm(BaseForm):
    """
    Geoh5py group uijson form.
    """

    value: OptionalUUID
    group_type: GroupTypes


class Association(str, Enum):
    """
    Geoh5py object association types.
    """

    VERTEX = "Vertex"
    CELL = "Cell"
    FACE = "Face"


class DataType(str, Enum):
    """
    Geoh5py data types.
    """

    INTEGER = "Integer"
    FLOAT = "Float"
    BOOLEAN = "Boolean"
    REFERENCED = "Referenced"
    VECTOR = "Vector"
    DATETIME = "DateTime"
    GEOMETRIC = "Geometric"
    TEXT = "Text"


UUIDOrNumber = Annotated[
    UUID | float | int | None,  # pylint: disable=unsupported-binary-operation
    BeforeValidator(empty_string_to_none),
    PlainSerializer(uuid_to_string_or_numeric),
]


class DataForm(BaseForm):
    """
    Geoh5py data uijson form.
    """

    value: UUIDOrNumber
    parent: str
    association: Association | list[Association]
    data_type: DataType | list[DataType]
    is_value: bool = False
    property: OptionalUUID = None
    min: float = -np.inf
    max: float = np.inf
    precision: int = 2

    @model_validator(mode="after")
    def value_if_is_value(self):
        if (
            "is_value" in self.model_fields_set  # pylint: disable=unsupported-membership-test
            and self.is_value
        ):
            if isinstance(self.value, UUID):
                raise ValueError("Value must be numeric if is_value is True.")
        return self

    @model_validator(mode="after")
    def property_if_not_is_value(self):
        if (
            "is_value" in self.model_fields_set  # pylint: disable=unsupported-membership-test
            and "property" not in self.model_fields_set  # pylint: disable=unsupported-membership-test
        ):
            raise ValueError("A property must be provided if is_value is used.")
        return self

    def flatten(self):
        """Returns the data for the form."""
        if (
            "is_value" in self.model_fields_set  # pylint: disable=unsupported-membership-test
            and not self.is_value
        ):
            return self.property
        return self.value
